@@ -9,12 +9,14 @@ import com.puppetlabs.http.client.RequestOptions;
 import com.puppetlabs.http.client.ResponseBodyType;
 import com.puppetlabs.http.client.impl.metrics.TimerUtils;
 
+import com.puppetlabs.ssl_utils.SSLUtils;
 import org.apache.commons.io.IOUtils;
 import org.apache.http.Consts;
 import org.apache.http.Header;
 import org.apache.http.HttpEntity;
 import org.apache.http.HttpRequest;
 import org.apache.http.HttpResponse;
+import org.apache.http.StatusLine;
 import org.apache.http.ParseException;
 import org.apache.http.ProtocolException;
 import org.apache.http.client.RedirectStrategy;
@@ -34,6 +36,7 @@ import org.apache.http.client.methods.HttpUriRequest;
 import org.apache.http.client.protocol.HttpClientContext;
 import org.apache.http.client.protocol.ResponseContentEncoding;
 import org.apache.http.concurrent.FutureCallback;
+import org.apache.http.conn.ssl.DefaultHostnameVerifier;
 import org.apache.http.entity.ContentType;
 import org.apache.http.entity.InputStreamEntity;
 import org.apache.http.impl.client.DefaultRedirectStrategy;
@@ -62,6 +65,7 @@ import java.net.URI;
 import java.nio.charset.UnsupportedCharsetException;
 import java.security.KeyManagementException;
 import java.security.NoSuchAlgorithmException;
+import java.security.NoSuchProviderException;
 import java.security.cert.CertificateException;
 import java.security.cert.X509Certificate;
 import java.util.HashMap;
@@ -214,23 +218,27 @@ public class JavaClient {
     }
 
     public static CoercedClientOptions coerceClientOptions(ClientOptions options) {
-        SSLContext sslContext = null;
+        final SSLContext sslContext;
         if (options.getSslContext() != null) {
             sslContext = options.getSslContext();
         } else if (options.getInsecure()) {
             sslContext = getInsecureSslContext();
+        } else {
+            sslContext = null;
         }
 
-        String[] sslProtocols = null;
+        final String[] sslProtocols;
         if (options.getSslProtocols() != null) {
             sslProtocols = options.getSslProtocols();
         } else {
             sslProtocols = ClientOptions.DEFAULT_SSL_PROTOCOLS;
         }
 
-        String[] sslCipherSuites = null;
+        final String[] sslCipherSuites;
         if (options.getSslCipherSuites() != null) {
             sslCipherSuites = options.getSslCipherSuites();
+        } else {
+            sslCipherSuites = null;
         }
 
         boolean forceRedirects = options.getForceRedirects();
@@ -245,10 +253,14 @@ public class JavaClient {
     }
 
     private static SSLContext getInsecureSslContext() {
-        SSLContext context = null;
+        final SSLContext context;
         try {
-            context = SSLContext.getInstance(PROTOCOL);
-        } catch (NoSuchAlgorithmException e) {
+            if (SSLUtils.isFIPS()) {
+                context = SSLContext.getInstance(SSLUtils.TLS_PROTOCOL, SSLUtils.BOUNCYCASTLE_JSSE_PROVIDER);
+            } else {
+                context = SSLContext.getInstance(PROTOCOL);
+            }
+        } catch (NoSuchAlgorithmException | NoSuchProviderException e) {
             throw new HttpClientException("Unable to construct HTTP context", e);
         }
         try {
@@ -297,16 +309,22 @@ public class JavaClient {
             ContentType contentType = null;
             String contentTypeValue = headers.get("content-type");
             if (contentTypeValue != null && !contentTypeValue.isEmpty()) {
-                contentType = ContentType.parse(contentTypeValue);
+                try {
+                    contentType = ContentType.parse(contentTypeValue);
+                } catch (ParseException e) {
+                    LOGGER.error("Unable to parse response content-type", e);
+                }
             }
             if (requestOptions.getAs() == ResponseBodyType.TEXT) {
                 body = coerceBodyType((InputStream) body, requestOptions.getAs(), contentType);
             }
+            StatusLine status = httpResponse.getStatusLine();
             responseDeliveryDelegate.deliverResponse(requestOptions,
                     origContentEncoding,
                     body,
                     headers,
-                    httpResponse.getStatusLine().getStatusCode(),
+                    status.getStatusCode(),
+                    status.getReasonPhrase(),
                     contentType,
                     callback);
         } catch (Exception e) {
@@ -512,12 +530,18 @@ public class JavaClient {
     public static CloseableHttpAsyncClient createClient(ClientOptions clientOptions) {
         CoercedClientOptions coercedOptions = coerceClientOptions(SslUtils.configureSsl(clientOptions));
         HttpAsyncClientBuilder clientBuilder = HttpAsyncClients.custom();
-        if (coercedOptions.getSslContext() != null) {
+        clientBuilder.setMaxConnPerRoute(clientOptions.getMaxConnectionsPerRoute());
+        clientBuilder.setMaxConnTotal(clientOptions.getMaxConnectionsTotal());
+
+        SSLContext context = coercedOptions.getSslContext();
+        if (context != null) {
+            // this requires an initialized SSLContext
+
             clientBuilder.setSSLStrategy(
-                    new SSLIOSessionStrategy(coercedOptions.getSslContext(),
+                    new SSLIOSessionStrategy(context,
                             coercedOptions.getSslProtocols(),
                             coercedOptions.getSslCipherSuites(),
-                            SSLIOSessionStrategy.BROWSER_COMPATIBLE_HOSTNAME_VERIFIER));
+                            new DefaultHostnameVerifier()));
         }
         RedirectStrategy redirectStrategy;
         if (!coercedOptions.getFollowRedirects()) {
