@@ -1,8 +1,9 @@
 (ns puppetlabs.http.client.async-plaintext-test
   (:import (com.puppetlabs.http.client Async RequestOptions ClientOptions)
            (org.apache.http.impl.nio.client HttpAsyncClients)
-           (java.net URI SocketTimeoutException ServerSocket)
-           (java.util Locale))
+           (java.net URI SocketTimeoutException ServerSocket URL)
+           (java.util Locale)
+           (java.util.concurrent CountDownLatch TimeUnit))
   (:require [clojure.test :refer :all]
             [puppetlabs.http.client.test-common :refer :all]
             [puppetlabs.i18n.core :as i18n]
@@ -64,96 +65,107 @@
 (deftest persistent-async-client-test
   (testlogging/with-test-logging
     (testutils/with-app-with-config app
-    [jetty9/jetty9-service test-web-service]
-    {:webserver {:port 10000}}
-    (testing "java async client"
-      (let [request-options (RequestOptions. (URI. "http://localhost:10000/hello/"))
-            client-options (ClientOptions.)
-            client (Async/createClient client-options)]
-        (testing "HEAD request with persistent async client"
-          (let [response (.head client request-options)]
-            (is (= 200 (.getStatus (.deref response))))
-            (is (= nil (.getBody (.deref response))))))
-        (testing "GET request with persistent async client"
-          (let [response (.get client request-options)]
-            (is (= 200 (.getStatus (.deref response))))
-            (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
-        (testing "POST request with persistent async client"
-          (let [response (.post client request-options)]
-            (is (= 200 (.getStatus (.deref response))))
-            (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
-        (testing "PUT request with persistent async client"
-          (let [response (.put client request-options)]
-            (is (= 200 (.getStatus (.deref response))))
-            (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
-        (testing "DELETE request with persistent async client"
-          (let [response (.delete client request-options)]
-            (is (= 200 (.getStatus (.deref response))))
-            (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
-        (testing "TRACE request with persistent async client"
-          (let [response (.trace client request-options)]
-            (is (= 200 (.getStatus (.deref response))))
-            (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
-        (testing "OPTIONS request with persistent async client"
-          (let [response (.options client request-options)]
-            (is (= 200 (.getStatus (.deref response))))
-            (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
-        (testing "PATCH request with persistent async client"
-          (let [response (.patch client request-options)]
-            (is (= 200 (.getStatus (.deref response))))
-            (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
-        (testing "client closes properly"
-          (.close client)
-          (is (thrown? IllegalStateException
-                       (.get client request-options))))))
-    (testing "clojure async client"
-      (let [client (async/create-client {})]
-        (testing "HEAD request with persistent async client"
-          (let [response (common/head client "http://localhost:10000/hello/")]
-            (is (= 200 (:status @response)))
-            (is (= nil (:body @response)))))
-        (testing "GET request with persistent async client"
-          (let [response (common/get client "http://localhost:10000/hello/")]
-            (is (= 200 (:status @response)))
-            (is (= "Hello, World!" (slurp (:body @response))))))
-        (testing "POST request with persistent async client"
-          (let [response (common/post client "http://localhost:10000/hello/")]
-            (is (= 200 (:status @response)))
-            (is (= "Hello, World!" (slurp (:body @response))))))
-        (testing "PUT request with persistent async client"
-          (let [response (common/put client "http://localhost:10000/hello/")]
-            (is (= 200 (:status @response)))
-            (is (= "Hello, World!" (slurp (:body @response))))))
-        (testing "DELETE request with persistent async client"
-          (let [response (common/delete client "http://localhost:10000/hello/")]
-            (is (= 200 (:status @response)))
-            (is (= "Hello, World!" (slurp (:body @response))))))
-        (testing "TRACE request with persistent async client"
-          (let [response (common/trace client "http://localhost:10000/hello/")]
-            (is (= 200 (:status @response)))
-            (is (= "Hello, World!" (slurp (:body @response))))))
-        (testing "OPTIONS request with persistent async client"
-          (let [response (common/options client "http://localhost:10000/hello/")]
-            (is (= 200 (:status @response)))
-            (is (= "Hello, World!" (slurp (:body @response))))))
-        (testing "PATCH request with persistent async client"
-          (let [response (common/patch client "http://localhost:10000/hello/")]
-            (is (= 200 (:status @response)))
-            (is (= "Hello, World!" (slurp (:body @response))))))
-        (testing "GET request via request function with persistent async client"
-          (let [response (common/make-request client "http://localhost:10000/hello/" :get)]
-            (is (= 200 (:status @response)))
-            (is (= "Hello, World!" (slurp (:body @response))))))
-        (testing "Bad verb request via request function with persistent async client"
-          (is (thrown? IllegalArgumentException
-                       (common/make-request client
-                                            "http://localhost:10000/hello/"
-                                            :bad))))
-        (testing "client closes properly"
-          (common/close client)
-          (is (thrown? IllegalStateException
-                       (common/get client
-                                   "http://localhost:10000/hello/")))))))))
+      [jetty9/jetty9-service test-web-service]
+      {:webserver {:port 10000}}
+      (testing "java async client"
+        (let [request-options (RequestOptions. (URI. "http://localhost:10000/hello/"))
+              client-options (ClientOptions.)
+              client (Async/createClient client-options)]
+          (testing "HEAD request to missing endpoint"
+            (let [response (.head client
+                                  (RequestOptions. (URI. "http://localhost:10000/missing")))]
+              (is (= 404 (.getStatus (.deref response))))
+              (is (= "Not Found" (.getReasonPhrase (.deref response))))))
+          (testing "HEAD request with persistent async client"
+            (let [response (.head client request-options)]
+              (is (= 200 (.getStatus (.deref response))))
+              (is (= "OK" (.getReasonPhrase (.deref response))))
+              (is (= nil (.getBody (.deref response))))))
+          (testing "GET request with persistent async client"
+            (let [response (.get client request-options)]
+              (is (= 200 (.getStatus (.deref response))))
+              (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
+          (testing "POST request with persistent async client"
+            (let [response (.post client request-options)]
+              (is (= 200 (.getStatus (.deref response))))
+              (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
+          (testing "PUT request with persistent async client"
+            (let [response (.put client request-options)]
+              (is (= 200 (.getStatus (.deref response))))
+              (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
+          (testing "DELETE request with persistent async client"
+            (let [response (.delete client request-options)]
+              (is (= 200 (.getStatus (.deref response))))
+              (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
+          (testing "TRACE request with persistent async client"
+            (let [response (.trace client request-options)]
+              (is (= 200 (.getStatus (.deref response))))
+              (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
+          (testing "OPTIONS request with persistent async client"
+            (let [response (.options client request-options)]
+              (is (= 200 (.getStatus (.deref response))))
+              (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
+          (testing "PATCH request with persistent async client"
+            (let [response (.patch client request-options)]
+              (is (= 200 (.getStatus (.deref response))))
+              (is (= "Hello, World!" (slurp (.getBody (.deref response)))))))
+          (testing "client closes properly"
+            (.close client)
+            (is (thrown? IllegalStateException
+                         (.get client request-options))))))
+      (testing "clojure async client"
+        (let [client (async/create-client {})]
+          (testing "HEAD request to missing endpoint"
+            (let [response (common/head client "http://localhost:10000/missing")]
+              (is (= 404 (:status @response)))
+              (is (= "Not Found" (:reason-phrase @response)))))
+          (testing "HEAD request with persistent async client"
+            (let [response (common/head client "http://localhost:10000/hello/")]
+              (is (= 200 (:status @response)))
+              (is (= "OK" (:reason-phrase @response)))
+              (is (= nil (:body @response)))))
+          (testing "GET request with persistent async client"
+            (let [response (common/get client "http://localhost:10000/hello/")]
+              (is (= 200 (:status @response)))
+              (is (= "Hello, World!" (slurp (:body @response))))))
+          (testing "POST request with persistent async client"
+            (let [response (common/post client "http://localhost:10000/hello/")]
+              (is (= 200 (:status @response)))
+              (is (= "Hello, World!" (slurp (:body @response))))))
+          (testing "PUT request with persistent async client"
+            (let [response (common/put client "http://localhost:10000/hello/")]
+              (is (= 200 (:status @response)))
+              (is (= "Hello, World!" (slurp (:body @response))))))
+          (testing "DELETE request with persistent async client"
+            (let [response (common/delete client "http://localhost:10000/hello/")]
+              (is (= 200 (:status @response)))
+              (is (= "Hello, World!" (slurp (:body @response))))))
+          (testing "TRACE request with persistent async client"
+            (let [response (common/trace client "http://localhost:10000/hello/")]
+              (is (= 200 (:status @response)))
+              (is (= "Hello, World!" (slurp (:body @response))))))
+          (testing "OPTIONS request with persistent async client"
+            (let [response (common/options client "http://localhost:10000/hello/")]
+              (is (= 200 (:status @response)))
+              (is (= "Hello, World!" (slurp (:body @response))))))
+          (testing "PATCH request with persistent async client"
+            (let [response (common/patch client "http://localhost:10000/hello/")]
+              (is (= 200 (:status @response)))
+              (is (= "Hello, World!" (slurp (:body @response))))))
+          (testing "GET request via request function with persistent async client"
+            (let [response (common/make-request client "http://localhost:10000/hello/" :get)]
+              (is (= 200 (:status @response)))
+              (is (= "Hello, World!" (slurp (:body @response))))))
+          (testing "Bad verb request via request function with persistent async client"
+            (is (thrown? IllegalArgumentException
+                         (common/make-request client
+                                              "http://localhost:10000/hello/"
+                                              :bad))))
+          (testing "client closes properly"
+            (common/close client)
+            (is (thrown? IllegalStateException
+                         (common/get client
+                                     "http://localhost:10000/hello/")))))))))
 
 (deftest java-api-cookie-test
   (testlogging/with-test-logging
@@ -204,7 +216,7 @@
     (testutils/with-app-with-config app
       [jetty9/jetty9-service test-params-web-service]
       {:webserver {:port 8080}}
-        (testing "URL Query Parameters work with the Java client"
+      (testing "URI Query Parameters work with the Java client"
           (let [client (Async/createClient (ClientOptions.))]
             (try
               (let [request-options (RequestOptions.
@@ -215,7 +227,7 @@
                                                          (.deref response)))))))
               (finally
                 (.close client)))))
-        (testing "URL Query Parameters work with the clojure client"
+      (testing "string URL Query Parameters work with the clojure client"
           (with-open [client (async/create-client {})]
             (let [opts     {:method       :get
                             :url          "http://localhost:8080/params/"
@@ -224,14 +236,32 @@
                   response (common/get client "http://localhost:8080/params" opts)]
                 (is (= 200 (:status @response)))
                 (is (= queryparams (read-string (:body @response)))))))
-        (testing "URL Query Parameters can be set directly in the URL"
+      (testing "URL based Query Parameters work with the clojure client"
+        (with-open [client (async/create-client {})]
+          (let [opts {:method       :get
+                      :url          (URL. "http://localhost:8080/params/")
+                      :query-params queryparams
+                      :as           :text}
+                response (common/get client "http://localhost:8080/params" opts)]
+            (is (= 200 (:status @response)))
+            (is (= queryparams (read-string (:body @response)))))))
+      (testing "URI based Query Parameters work with the clojure client"
+        (with-open [client (async/create-client {})]
+          (let [opts     {:method       :get
+                          :url          (.toURI (URL. "http://localhost:8080/params/"))
+                          :query-params queryparams
+                          :as           :text}
+                response (common/get client "http://localhost:8080/params" opts)]
+            (is (= 200 (:status @response)))
+            (is (= queryparams (read-string (:body @response)))))))
+      (testing "string URL Query Parameters can be set directly in the URL"
           (with-open [client (async/create-client {})]
             (let [response (common/get client
                                        "http://localhost:8080/params?paramone=one"
                                        {:as :text})]
               (is (= 200 (:status @response)))
               (is (= (str {"paramone" "one"}) (:body @response))))))
-        (testing (str "URL Query Parameters set in URL are overwritten if params "
+      (testing (str "string URL Query Parameters set in URL are overwritten if params "
                       "are also specified in options map")
           (with-open [client (async/create-client {})]
             (let [response (common/get client
@@ -425,8 +455,10 @@
                 (is (= "Hello, World!" (:body response)))))))))))
 
 (deftest empty-content-type-async
-  (testing "content-type parsing handles empty content-type"
+  (testing "content-type parsing handles invalid content-type"
     (testlogging/with-test-logging
+     ;; while the content-type is empty, the request actually gets sent
+     ;; as ";<character-set>", which is invalid
       (testwebserver/with-test-webserver app-with-empty-content-type port
          (let [url (str "http://localhost:" port "/hello")]
            (testing "java persistent async client"
@@ -452,3 +484,96 @@
                 (let [response @(common/get client url {:as :text})]
                   (is (= 200 (:status response)))
                   (is (= "es-ES" (:body response))))))))))))
+
+(deftest client-route-and-total-limits
+  (testing "client limits 2 requests per route by default"
+    (let [actual-count (atom 0)
+          countdown (CountDownLatch. 3)
+          fake-app (fn [_]
+                     (swap! actual-count inc)
+                     (.countDown countdown)
+                     (.await countdown)
+                     {:status 200
+                      :body "Hello, World!"})]
+      (testlogging/with-test-logging
+        (testwebserver/with-test-webserver fake-app port
+          (let [url (str "http://localhost:" port "/hello")]
+            (testing "clojure persistent async client"
+              (with-open [client (async/create-client {})]
+                (dotimes [n 10] (future (common/get client url {:as :text})))
+                (is (= false (.await countdown 5 TimeUnit/SECONDS)))
+                (is (= 2 @actual-count)))))))))
+
+  (testing "passing client route limit of 0 selects default behavior (a limit of 2)"
+    (let [actual-count (atom 0)
+          countdown (CountDownLatch. 3)
+          fake-app (fn [_]
+                     (swap! actual-count inc)
+                     (.countDown countdown)
+                     (.await countdown)
+                     {:status 200
+                      :body "Hello, World!"})]
+      (testlogging/with-test-logging
+       (testwebserver/with-test-webserver fake-app port
+         (let [url (str "http://localhost:" port "/hello")]
+           (testing "clojure persistent async client"
+             (with-open [client (async/create-client {:max-connections-per-route 0})]
+               (dotimes [n 10] (future (common/get client url {:as :text})))
+               (is (= false (.await countdown 5 TimeUnit/SECONDS)))
+               (is (= 2 @actual-count)))))))))
+
+
+  (testing "client limits specified requests per route"
+    (let [actual-count (atom 0)
+          countdown (CountDownLatch. 4)
+          fake-app (fn [_]
+                     (swap! actual-count inc)
+                     (.countDown countdown)
+                     (.await countdown)
+                     {:status 200
+                      :body "Hello, World!"})]
+      (testlogging/with-test-logging
+       (testwebserver/with-test-webserver fake-app port
+         (let [url (str "http://localhost:" port "/hello")]
+           (testing "clojure persistent async client"
+             (with-open [client (async/create-client {:max-connections-per-route 3})]
+               (dotimes [n 10] (future (common/get client url {:as :text})))
+               (is (= false (.await countdown 5 TimeUnit/SECONDS)))
+               (is (= 3 @actual-count)))))))))
+
+  (testing "client route limit of 11 does not limit requests per route when less than 11"
+    (let [actual-count (atom 0)
+          countdown (CountDownLatch. 10)
+          fake-app (fn [_]
+                     (swap! actual-count inc)
+                     (.countDown countdown)
+                     (.await countdown)
+                     {:status 200
+                      :body "Hello, World!"})]
+      (testlogging/with-test-logging
+       (testwebserver/with-test-webserver fake-app port
+         (let [url (str "http://localhost:" port "/hello")]
+           (testing "clojure persistent async client"
+             (with-open [client (async/create-client {:max-connections-per-route 11})]
+               (dotimes [n 10] (future (common/get client url {:as :text})))
+               (is (= true (.await countdown 5 TimeUnit/SECONDS)))
+               (is (= 10 @actual-count)))))))))
+
+  (testing "overall limit applies"
+    (let [actual-count (atom 0)
+          countdown (CountDownLatch. 4)
+          fake-app (fn [_]
+                     (swap! actual-count inc)
+                     (.countDown countdown)
+                     (.await countdown)
+                     {:status 200
+                      :body "Hello, World!"})]
+      (testlogging/with-test-logging
+       (testwebserver/with-test-webserver fake-app port
+         (let [url (str "http://localhost:" port "/hello")]
+           (testing "clojure persistent async client"
+             (with-open [client (async/create-client {:max-connections-per-route 11
+                                                      :max-connections-total 3})]
+               (dotimes [n 10] (future (common/get client url {:as :text})))
+               (is (= false (.await countdown 5 TimeUnit/SECONDS)))
+               (is (= 3 @actual-count))))))))))
